-- -*- Mode: LUA; tab-width: 2 -*-

-- wbgen2 - a simple Wishbone slave generator
-- (c) 2010 Tomasz Wlostowski
-- CERN BE-Co-HT
-- LICENSED UNDER GPL v2

function gen_vhdl_field_prefix(field, reg)
	local field_count;

	if(reg.hdl_prefix == nil) then
		die("no prefix specified for reg: "..reg.name);
	end

	field_count = 0;
	foreach_subfield(reg, function(field, reg) field_count = field_count+1; end );

	if(field.count == 0) then
		die("empty reg: "..reg.name);
	end
		
	if(field.prefix == nil) then
		if(field_count >1 ) then die("multiple anonymous-prefix fields declared for reg: "..reg.name); end
		return  string.lower(periph.hdl_prefix.."_"..reg.hdl_prefix);
	end

	return  string.lower(periph.hdl_prefix.."_"..reg.hdl_prefix.."_"..field.hdl_prefix);

end

-- generates VHDL for monostable-type field (both same-clock and other-clock)
function gen_vhdl_code_monostable(field, reg)
	local prefix = gen_vhdl_field_prefix(field, reg); 

--	field.prefix = prefix;

-- monostable type field using WB bus clock
  if(field.clock == nil) then
-- WB-synchronous monostable port (bus write-only)
		field.signals = 	{	signal(BIT, 0, prefix.."_dly0"),
												signal(BIT, 0, prefix.."_int") };
					
		field.ports =		{	port(BIT, 0, "out", prefix.."_o", "Port for MONOSTABLE field: '"..field.name.."' in reg: '"..reg.name.."'" ) };
		field.acklen = 3;

		field.extra_code = {"process (wb_clk_i, rst_n_i)",
												"begin",
												"if(rst_n_i = '0') then",
												prefix.."_dly0 <= '0';",
						      			prefix.."_o <= '0';",
												"elsif rising_edge(wb_clk_i) then",
												prefix.."_dly0 <= "..prefix.."_int;",
												prefix.."_o <= "..prefix.."_int and (not "..prefix.."_dly0);",
												"end if;",
												"end process;" };
	
  	field.reset_code_main =  	{ prefix.."_int <= '0';" };
	  field.write_code = 				{ prefix.."_int <= wb_data_i("..field.offset..");" };
	  field.read_code = 				{ "" };
	  field.ackgen_code = 			{ prefix.."_int <= '0';" };

  else
-- WB-asynchronous monostable port (bus write-only)	
 		field.signals = 					{	signal(BIT, 0, prefix.."_int"), 
 																signal(BIT, 0, prefix.."_int_delay"), 
																signal(BIT, 0, prefix.."_sync0"), 
																signal(BIT, 0, prefix.."_sync1"),
																signal(BIT, 0, prefix.."_sync2") };
					
	  field.ports = 						{	port(BIT, 0, "out", prefix.."_o", "Port for asynchronous (clock: "..field.clock..") MONOSTABLE field: '"..field.name.."' in reg: '"..reg.name.."'") };

	  field.acklen = 						5;
	   
	  field.extra_code =				{	"process ("..field.clock..", rst_n_i)",
																"begin",
																"if(rst_n_i = '0') then",
																prefix.."_o <= '0';",
																prefix.."_sync0 <= '0';",
																prefix.."_sync1 <= '0';",
																prefix.."_sync2 <= '0';",
																"elsif rising_edge("..field.clock..") then",
																prefix.."_sync0 <= "..prefix.."_int;",
																prefix.."_sync1 <= "..prefix.."_sync0;",
																prefix.."_sync2 <= "..prefix.."_sync1;",
																prefix.."_o <= "..prefix.."_sync2 and (not "..prefix.."_sync1);",
																"end if;",
																"end process;" };
	
	  field.reset_code_main = 	{	prefix.."_int <= '0';",
	  														prefix.."_int_delay <= '0';" };

	  field.write_code =				{ prefix.."_int <= wb_data_i("..field.offset..");",
	  														prefix.."_int_delay <= wb_data_i("..field.offset..");" };

	  field.read_code =					{ 	"" };

	  field.ackgen_code_pre =		{ prefix.."_int <= "..prefix.."_int_delay;",
	  														prefix.."_int_delay <= '0';" };
  end
end


-- generates code for BIT-type field
function gen_vhdl_code_bit(field, reg)
	local prefix = gen_vhdl_field_prefix(field, reg); 

  field.prefix = prefix;

-- BIT-type field using WB bus clock
  if(field.clock == nil) then
    if(field.access == ACC_RW_RO) then
-- bus(read-write), dev(read-only) bitfield
			field.ports =						{ port(BIT, 0, "out", prefix.."_o", "Port for BIT field: '"..field.name.."' in reg: '"..reg.name.."'" ) };
			field.signals = 				{ signal(BIT, 0, prefix.."_int") };
			field.acklen = 					1;
			field.write_code =			{ prefix.."_int <= wb_data_i("..field.offset..");" };
			field.read_code = 			{ "wb_data_out_int("..field.offset..") <= "..prefix.."_int;" };
	    field.reset_code_main =	{ prefix.."_int <= '0';" };
			field.extra_code =			{ prefix.."_o <= "..prefix.."_int; "};

	  elseif (field.access == ACC_RO_WO) then
-- bus(read-only), dev(read-only) bitfield
	   	field.ports =						{ port(BIT, 0, "in", prefix.."_i", "Port for BIT field: '"..field.name.."' in reg: '"..reg.name.."'") };
			field.signals = 				{  };
			field.acklen = 					1;
			field.write_code =			{  };
			field.read_code = 			{ "wb_data_out_int("..field.offset..") <= "..prefix.."_i;" };
      field.reset_code_main =	{  };
			field.extra_code =	{  };

	  elseif (field.access == ACC_WO_RO) then
-- bus(write-only), dev(read-only) bitfield - unsupported yet (use RW/RO type instead)
			die("WO-RO type unsupported yet ("..field.name..")");

	  elseif (field.access == ACC_RW_RW) then
-- dual-write bitfield (both from the bus and the device)
			if(field.load == LOAD_EXT) then

-- external load type (e.g. the register itself is placed outside the WB slave, which only outputs new value and asserts the "load" signal for single clock cycle upon bus write.
		    field.ports =						{	port(BIT, 0, "out", prefix.."_o", "Ports for BIT field: '"..field.name.."' in reg: '"..reg.name.."'"),
							  									port(BIT, 0, "in", prefix.."_i"),	    
																	port(BIT, 0, "out", prefix.."_load_o") };	    
	
		    field.acklen =			 		1;
		    
		    field.read_code = 			{ "wb_data_out_int("..field.offset..") <= "..prefix.."_i;" };
		    field.write_code = 			{ prefix.."_load_o <= '1';" };
		    field.extra_code =			{ prefix.."_o <= wb_data_i("..field.offset..");" };
		    field.ackgen_code_pre = { prefix.."_load_o <= '0';" };
		    field.ackgen_code  = 		{ prefix.."_load_o <= '0';" };
		    field.reset_code_main = { prefix.."_load_o <= '0';" };
	
			else
		    die("internal RW/RW register storage unsupported yet ("..field.name..")");
			end
    end
	else
-- asynchronous bit-type register

    if(field.access == ACC_RW_RO) then
-- bus(read-write), dev(read-only) bitfield, asynchronous
			field.ports =						{ port(BIT, 0, "out", prefix.."_o", "Port for asynchronous (clock: "..field.clock..") BIT field: '"..field.name.."' in reg: '"..reg.name.."'") };
			field.signals = 				{ signal(BIT, 0, prefix.."_int"),
												  			signal(BIT, 0, prefix.."_sync0"),
												  			signal(BIT, 0, prefix.."_sync1") };
					
			field.acklen = 					4;
			field.write_code =			{ prefix.."_int <= wb_data_i("..field.offset..");" };
			field.read_code = 			{ "wb_data_out_int("..field.offset..") <= "..prefix.."_int;" };

	    field.reset_code_main =	{ prefix.."_int <= '0';" };

			field.extra_code =			{	"-- synchronizer chain for field : "..field.name.." (type RW/RO, wb_clk_i <-> "..field.clock..")",
																"process ("..field.clock..", rst_n_i)",
																"begin",
																"if(rst_n_i = '0') then",
								    						prefix.."_o <= '0';",
							    							prefix.."_sync0 <= '0';",
																prefix.."_sync1 <= '0';",
																"elsif rising_edge("..field.clock..") then",
																prefix.."_sync0 <= "..prefix.."_int;",
																prefix.."_sync1 <= "..prefix.."_sync0;",
																prefix.."_o <= "..prefix.."_sync1;",
																"end if;",
																"end process;" };

		elseif (field.access == ACC_RO_WO) then
-- bus(read-only), dev(write-only) bitfield, asynchronous

		field.ports =							{ port(BIT, 0, "in", prefix.."_i", "Port for asynchronous (clock: "..field.clock..") BIT field: '"..field.name.."' in reg: '"..reg.name.."'") };
		field.signals = 					{ signal(BIT, 0, prefix.."_sync0"),
															  signal(BIT, 0, prefix.."_sync1") };
					
		field.acklen = 						1;
		field.write_code =				{  };
		field.read_code = 				{ "wb_data_out_int("..field.offset..") <= "..prefix.."_sync1;" };

    field.reset_code_main =		{ };

		field.extra_code =				{	"-- synchronizer chain for field : "..field.name.." (type RO/WO, "..field.clock.." -> wb_clk_i)",
																"process (wb_clk_i, rst_n_i)",
																"begin",
																"if(rst_n_i = '0') then",
    														prefix.."_sync0 <= '0';",
																prefix.."_sync1 <= '0';",
				      						  		"elsif rising_edge(wb_clk_i) then",
										        		prefix.."_sync0 <= "..prefix.."_i;",
																prefix.."_sync1 <= "..prefix.."_sync0;",
																"end if;",
																"end process;" };

    elseif (field.access == ACC_RW_RW) then
-- asynchronous dual-write bitfield. Tough shit :/

		if(field.load ~= LOAD_EXT) then
		    die("Only external load is supported for RW/RW bit fields");
			end

			local comment = "Ports for asynchronous (clock: "..field.clock..") RW/RW BIT field: '"..field.name.."' in reg: '"..reg.name.."'";

   		field.ports =									{	port(BIT, 0, "out", prefix.."_o", comment),
																			port(BIT, 0, "in", 	prefix.."_i"),
																			port(BIT, 0, "out", prefix.."_load_o") };  


			field.signals = 							{ signal(BIT, 0, prefix.."_int_read"),
																			signal(BIT, 0, prefix.."_int_write"),
																			signal(BIT, 0, prefix.."_lw"),
																			signal(BIT, 0, prefix.."_lw_delay"),
																			signal(BIT, 0, prefix.."_lw_read_in_progress"),
																			signal(BIT, 0, prefix.."_lw_s0"),
																			signal(BIT, 0, prefix.."_lw_s1"),
																			signal(BIT, 0, prefix.."_lw_s2"),
																			signal(BIT, 0, prefix.."_rwsel") };
																			
			field.acklen = 									6;
		
			field.write_code =						{ prefix.."_int_write <= wb_data_i("..field.offset..");",
																			prefix.."_lw <= '1';",
																			prefix.."_lw_delay <= '1';",
																			prefix.."_lw_read_in_progress <= '0';",
																			prefix.."_rwsel <= '1'; " }; 

			field.read_code =						{ 	prefix.."_lw <= '1';",
																			prefix.."_lw_delay <= '1';",
																			prefix.."_lw_read_in_progress <= '1';",
																			prefix.."_rwsel <= '0'; " }; 
																			

			field.reset_code_main =				{ prefix.."_lw <= '0';",
																			prefix.."_lw_delay <= '0';",
																			prefix.."_lw_read_in_progress <= '0';",
																			prefix.."_rwsel <= '0';",
																			prefix.."_int_write <= '0';"
																		};
							

			field.ackgen_code_pre =				{ prefix.."_lw <= "..prefix.."_lw_delay; ",
																			prefix.."_lw_delay <= '0'; ",
																			"if ((ack_cntr = "..gen_vhdl_bin_literal(1,4)..") and ("..prefix.."_lw_read_in_progress = '1')) then",
																			"wb_data_out_int("..field.offset..") <= "..prefix.."_int_read;",
																			prefix.."_lw_read_in_progress <= '0';",
																			"end if;" };



			field.extra_code =						{ "-- asynchronous BIT register : "..field.name.." (type RW/WO, "..field.clock.." <-> wb_clk_i)",
											    					  "process ("..field.clock..", rst_n_i)",
													 					  "begin",
																			"if(rst_n_i = '0') then",
													 					  prefix.."_lw_s0 <= '0';",
																			prefix.."_lw_s1 <= '0';",
																			prefix.."_lw_s2 <= '0';",
																			prefix.."_int_read <= '0';",
																			prefix.."_load_o <= '0';",
																			prefix.."_o <= '0';",
													 	         "elsif rising_edge("..field.clock..") then",
																			prefix.."_lw_s0 <= "..prefix.."_lw;",
																			prefix.."_lw_s1 <= "..prefix.."_lw_s0;",
																			prefix.."_lw_s2 <= "..prefix.."_lw_s1;",
																			"if("..prefix.."_lw_s2 = '0' and "..prefix.."_lw_s1 = '1') then",
																			"if("..prefix.."_rwsel = '1') then",
																			prefix.."_o <= "..prefix.."_int_write;",
																			prefix.."_load_o <= '1';",
																			"else",
																			prefix.."_load_o <= '0';",
																			prefix.."_int_read <= "..prefix.."_i;",
																			"end if;",
																			"else",
																			prefix.."_load_o <= '0';",
																			"end if;",
																			"end if;",
																			"end process;" };

		elseif (field.access == ACC_WO_RO) then
			die("WO-RO type unsupported yet ("..field.name..")");
    end
	end
end

-- generates the bit-range for accessing a certain register field from WB-bus
function gen_vhdl_subrange(field)
	local s;
	s="("..(field.offset+field.size-1).." downto "..(field.offset)..")";
	return s;
end

-- converts field of type (slv) to type (type)
function convert_from_slv(type, prefix)
	if(type == UNSIGNED) then
		return "unsigned("..prefix..")";
	elseif (type == SIGNED) then
		return "signed("..prefix..")";
	else return prefix; end
end

-- ... and vice versa
function convert_to_slv(type, prefix)
	if(type == UNSIGNED) then
		return "std_logic_vector("..prefix..")";
	elseif (type == SIGNED) then
		return "std_logic_vector("..prefix..")";
	else return prefix; end
end

-- generates code for slv, signed or unsigned fields
function gen_vhdl_code_slv(field, reg)
	local prefix = gen_vhdl_field_prefix(field, reg); 

	local is_slv = (field.type == SLV);

	local name_slv_i = csel(is_slv , prefix.."_i", prefix.."_i_slv");
	local name_slv_o = csel(is_slv , prefix.."_o", prefix.."_o_slv");	

  field.prefix = prefix;

-- synchronous signed/unsigned/slv field
  if(field.clock == nil) then
  
	  if(field.access == ACC_RW_RO) then
-- bus(read-write), dev(read-only) slv
			field.ports =						{ port(field.type, field.size, "out", prefix.."_o", "Port for "..fieldtype_2_vhdl[field.type].." field: '"..field.name.."' in reg: '"..reg.name.."'") };
			field.signals = 				{ signal(SLV, field.size, prefix.."_int") };
			field.acklen = 					1;
			field.write_code =			{ prefix.."_int <= wb_data_i"..gen_vhdl_subrange(field)..";" };
			field.read_code = 			{ "wb_data_out_int"..gen_vhdl_subrange(field).." <= "..prefix.."_int;" };
	    field.reset_code_main =	{ prefix.."_int <= (others => '0');" };
			field.extra_code =			{ prefix.."_o <= "..convert_from_slv(field.type, prefix.."_int")..";"};
		
    elseif (field.access == ACC_RO_WO) then
-- bus(read-only), dev(write-only) slv
			field.ports =						{ port(field.type, field.size, "in", prefix.."_i",  "Port for "..fieldtype_2_vhdl[field.type].." field: '"..field.name.."' in reg: '"..reg.name.."'") };
			field.signals = 				{  };
			field.acklen = 					1;
			field.write_code =			{  };
			field.read_code = 			{ "wb_data_out_int"..gen_vhdl_subrange(field).." <= "..convert_to_slv(field.type, prefix.."_i")..";" };
   		field.reset_code_main =	{  };
			field.extra_code =			{  };

	  elseif (field.access == ACC_RW_RW) then
-- bus(read-write), dev(read-write) slv
   		if(field.load ~= LOAD_EXT) then
		    die("Only external load is supported for RW/RW slv/signed/unsigned fields");
			end

		    field.ports =						{	port(field.type, field.size, "out", prefix.."_o",  "Port for "..fieldtype_2_vhdl[field.type].." field: '"..field.name.."' in reg: '"..reg.name.."'"),
										    					port(field.type, field.size, "in", prefix.."_i"),	    
																	port(BIT, 0, "out", prefix.."_load_o") };	    
	
		    field.acklen = 					1;
		    
		    field.read_code = 			{ "wb_data_out_int"..gen_vhdl_subrange(field).." <= "..convert_to_slv(field.type, prefix.."_i")..";" };
		    field.write_code = 			{ prefix.."_load_o <= '1';" };
		    field.extra_code =			{ prefix.."_o <= "..convert_from_slv(field.type, "wb_data_i"..gen_vhdl_subrange(field))..";" };
		    field.ackgen_code_pre = { prefix.."_load_o <= '0';" };
		    field.ackgen_code = 		{ prefix.."_load_o <= '0';" };
		    field.reset_code_main = { prefix.."_load_o <= '0';" };
		end
	else
-- asynchronous register. Even tougher shit :(

		if(field.access == ACC_RW_RO) then
-- bus(read-write), dev(read-only) slv/signed/unsigned
			local comment = "Port for asynchronous (clock: "..field.clock..") "..fieldtype_2_vhdl[field.type].." field: '"..field.name.."' in reg: '"..reg.name.."'";

			field.ports =									{ port(field.type, field.size, "out", prefix.."_o", comment) };

			field.signals = 							{ signal(SLV, field.size, prefix.."_int"),
																			signal(BIT, 0, prefix.."_swb"),
																			signal(BIT, 0, prefix.."_swb_delay"),
																			signal(BIT, 0, prefix.."_swb_s0"),
																			signal(BIT, 0, prefix.."_swb_s1"),
																			signal(BIT, 0, prefix.."_swb_s2") };
							
			field.acklen = 								4;
		
			field.write_code =						{ prefix.."_int <= wb_data_i"..gen_vhdl_subrange(field)..";",
																			prefix.."_swb <= '1';",
																			prefix.."_swb_delay <= '1';" };
							
			field.read_code = 						{ "wb_data_out_int"..gen_vhdl_subrange(field).." <= "..prefix.."_int;" };
		
			field.reset_code_main =				{ prefix.."_int <= (others => '0');",
																			prefix.."_swb <= '0'; ",
																			prefix.."_swb_delay <= '0';"  };
										
			field.ackgen_code_pre =				{ prefix.."_swb <= "..prefix.."_swb_delay; ",
																			prefix.."_swb_delay <= '0';" };
			

			field.extra_code =						{ "-- asynchronous "..fieldtype_2_vhdl[field.type].." register : "..field.name.." (type RW/RO, "..field.clock.." <-> wb_clk_i)",
																			"process ("..field.clock..", rst_n_i)",
																			"begin",
																			"if(rst_n_i = '0') then",
																			prefix.."_swb_s0 <= '0';",
																			prefix.."_swb_s1 <= '0';",
																			prefix.."_swb_s2 <= '0';",
																			prefix.."_o <= (others => '0');",
																			"elsif rising_edge("..field.clock..") then",
																			prefix.."_swb_s0 <= "..prefix.."_swb;",
																			prefix.."_swb_s1 <= "..prefix.."_swb_s0;",
																			prefix.."_swb_s2 <= "..prefix.."_swb_s1;",
																			"if("..prefix.."_swb_s2 = '0' and "..prefix.."_swb_s1 = '1') then",
																			prefix.."_o <= "..convert_from_slv(field.type, prefix.."_int")..";",
																			"end if;",
																			"end if;",
																			"end process;" };
																		
		elseif(field.access == ACC_RO_WO) then
-- bus(read-write), dev(read-only) slv
			local comment = "Port for asynchronous (clock: "..field.clock..") "..fieldtype_2_vhdl[field.type].." field: '"..field.name.."' in reg: '"..reg.name.."'";
	
			field.ports =									{ port(field.type, field.size, "in", prefix.."_i", comment) };

			field.signals = 							{ signal(SLV, field.size, prefix.."_int"),
																			signal(BIT, 0, prefix.."_lwb"),
																			signal(BIT, 0, prefix.."_lwb_delay"),
																			signal(BIT, 0, prefix.."_lwb_in_progress"),
																			signal(BIT, 0, prefix.."_lwb_s0"),
																			signal(BIT, 0, prefix.."_lwb_s1"),
																			signal(BIT, 0, prefix.."_lwb_s2") };
					  
			field.acklen = 								6;
			field.write_code =						{  }; 

			field.read_code = 						{ prefix.."_lwb <= '1';",
																			prefix.."_lwb_delay <= '1';",
																			prefix.."_lwb_in_progress <= '1';" };

			field.reset_code_main =				{ prefix.."_lwb <= '0';",
																			prefix.."_lwb_delay <= '0';",
																			prefix.."_lwb_in_progress <= '0';"  };


			field.ackgen_code_pre =				{ prefix.."_lwb <= "..prefix.."_lwb_delay; ",
																			prefix.."_lwb_delay <= '0'; ",
																			"if ((ack_cntr = "..gen_vhdl_bin_literal(1,4)..") and ("..prefix.."_lwb_in_progress = '1')) then",
																			"wb_data_out_int"..gen_vhdl_subrange(field).." <= "..prefix.."_int;",
																			prefix.."_lwb_in_progress <= '0';",
																			"end if;", };

			field.extra_code =						{ "-- asynchronous "..fieldtype_2_vhdl[field.type].." register : "..field.name.." (type RO/WO, "..field.clock.." <-> wb_clk_i)",
																			"process ("..field.clock..", rst_n_i)",
																			"begin",
																			"if(rst_n_i = '0') then",
																			prefix.."_lwb_s0 <= '0';",
																			prefix.."_lwb_s1 <= '0';",
																			prefix.."_lwb_s2 <= '0';",
																			prefix.."_int <= (others => '0');",
																			"elsif rising_edge("..field.clock..") then",
																			prefix.."_lwb_s0 <= "..prefix.."_lwb;",
																			prefix.."_lwb_s1 <= "..prefix.."_lwb_s0;",
																			prefix.."_lwb_s2 <= "..prefix.."_lwb_s1;",
																			"if("..prefix.."_lwb_s1 = '1' and "..prefix.."_lwb_s2 = '0') then",
																			prefix.."_int <= "..convert_to_slv(field.type, prefix.."_i")..";",
																			"end if;",
																			"end if;",
																			"end process;" };
	
		elseif(field.access == ACC_RW_RW) then
-- async bus(read-write), dev(read-write) slv. gooosh...

   		if(field.load ~= LOAD_EXT) then
		    die("Only external load is supported for RW/RW slv/signed/unsigned fields");
			end

			local comment = "Ports for asynchronous (clock: "..field.clock..") "..fieldtype_2_vhdl[field.type].." field: '"..field.name.."' in reg: '"..reg.name.."'";

   		field.ports =									{	port(field.type, field.size, "out", prefix.."_o", comment),
																			port(field.type, field.size, "in", prefix.."_i"),
																			port(BIT, 0, "out", prefix.."_load_o") };  


			field.signals = 							{ signal(SLV, field.size, prefix.."_int_read"),
																			signal(SLV, field.size, prefix.."_int_write"),
																			signal(BIT, 0, prefix.."_lw"),
																			signal(BIT, 0, prefix.."_lw_delay"),
																			signal(BIT, 0, prefix.."_lw_read_in_progress"),
																			signal(BIT, 0, prefix.."_lw_s0"),
																			signal(BIT, 0, prefix.."_lw_s1"),
																			signal(BIT, 0, prefix.."_lw_s2"),
																			signal(BIT, 0, prefix.."_rwsel") };
																			
							
			field.acklen = 									6;
		
			field.write_code =						{ prefix.."_int_write <= wb_data_i"..gen_vhdl_subrange(field)..";",
																			prefix.."_lw <= '1';",
																			prefix.."_lw_delay <= '1';",
																			prefix.."_lw_read_in_progress <= '0';",
																			prefix.."_rwsel <= '1'; " }; 

			field.read_code =						{ 	prefix.."_lw <= '1';",
																			prefix.."_lw_delay <= '1';",
																			prefix.."_lw_read_in_progress <= '1';",
																			prefix.."_rwsel <= '0'; " }; 
																			

			field.reset_code_main =				{ prefix.."_lw <= '0';",
																			prefix.."_lw_delay <= '0';",
																			prefix.."_lw_read_in_progress <= '0';",
																			prefix.."_rwsel <= '0';",
																			prefix.."_int_write <= (others => '0');"
																		};
							

			field.ackgen_code_pre =				{ prefix.."_lw <= "..prefix.."_lw_delay; ",
																			prefix.."_lw_delay <= '0'; ",
																			"if ((ack_cntr = "..gen_vhdl_bin_literal(1,4)..") and ("..prefix.."_lw_read_in_progress = '1')) then",
																			"wb_data_out_int"..gen_vhdl_subrange(field).." <= "..prefix.."_int_read;",
																			prefix.."_lw_read_in_progress <= '0';",
																			"end if;", };



			field.extra_code =						{ "-- asynchronous "..fieldtype_2_vhdl[field.type].." register : "..field.name.." (type RW/WO, "..field.clock.." <-> wb_clk_i)",
											    					  "process ("..field.clock..", rst_n_i)",
													 					  "begin",
																			"if(rst_n_i = '0') then",
													 					  prefix.."_lw_s0 <= '0';",
																			prefix.."_lw_s1 <= '0';",
																			prefix.."_lw_s2 <= '0';",
																			prefix.."_o <= (others => '0');",
																			prefix.."_load_o <= '0';",
 																			prefix.."_int_read <= (others => '0');",
													 	         "elsif rising_edge("..field.clock..") then",
																			prefix.."_lw_s0 <= "..prefix.."_lw;",
																			prefix.."_lw_s1 <= "..prefix.."_lw_s0;",
																			prefix.."_lw_s2 <= "..prefix.."_lw_s1;",
																			"if("..prefix.."_lw_s2 = '0' and "..prefix.."_lw_s1 = '1') then",
																			"if("..prefix.."_rwsel = '1') then",
																			prefix.."_o <= "..convert_from_slv(field.type, prefix.."_int_write")..";",
																			prefix.."_load_o <= '1';",
																			"else",
																			prefix.."_load_o <= '0';",
																			prefix.."_int_read <= "..convert_to_slv(field.type, prefix.."_i")..";",
																			"end if;",
																			"else",
																			prefix.."_load_o <= '0';",
																			"end if;",
																			"end if;",
																			"end process;" };
		end 
	end
end

function gen_vhdl_code_passthrough(field, reg)
	local prefix = gen_vhdl_field_prefix(field, reg); 

	if(field.clock == nil) then
-- sync pass-through

			local comment = "Ports for PASS_THROUGH field: '"..field.name.."' in reg: '"..reg.name.."'";

   		field.ports =									{	port(SLV, field.size, "out", prefix.."_o", comment),
   																		port(BIT, 0, "out", prefix.."_wr_o") };

			field.acklen = 1;
			
			field.reset_code_main	= 			{ prefix.."_wr_o <= '0';" };
			field.read_code = 						{};
			field.write_code = 						{ prefix.."_wr_o <= '1';" };
			field.ackgen_code_pre =				{ prefix.."_wr_o <= '0';" };
			field.ackgen_code	 =					{ prefix.."_wr_o <= '0';" };
			field.extra_code =						{ "-- pass-through field: "..field.name.." in register: "..reg.name,
																			prefix.."_o <= wb_data_i"..gen_vhdl_subrange(field)..";" }
	else

			local comment = "Ports for asynchronous (clock: "..field.clock..") PASS_THROUGH field: '"..field.name.."' in reg: '"..reg.name.."'";

   		field.ports =									{	port(SLV, field.size, "out", prefix.."_o", comment),
   																		port(BIT, 0, "out", prefix.."_wr_o") };

			field.signals =								{ signal(BIT, 0, prefix.."_wr_int"), 
																			signal(BIT, 0, prefix.."_wr_int_delay"), 
																			signal(BIT, 0, prefix.."_wr_sync0"), 
																			signal(BIT, 0, prefix.."_wr_sync1"),
																			signal(BIT, 0, prefix.."_wr_sync2") };

			field.acklen = 4;
			
			field.reset_code_main	= 			{ prefix.."_wr_int <= '0';",
																			prefix.."_wr_int_delay <= '0';" };
			field.read_code = 						{};

			field.write_code = 						{ prefix.."_wr_int <= '1';",
																			prefix.."_wr_int_delay <= '1';" };
			field.ackgen_code_pre =				{ prefix.."_wr_int <= "..prefix.."_wr_int_delay;",
																			prefix.."_wr_int_delay <= '0';" };

			field.extra_code =						{ "-- pass-through field: "..field.name.." in register: "..reg.name,
																			prefix.."_o <= wb_data_i"..gen_vhdl_subrange(field)..";",
																			"process ("..field.clock..", rst_n_i )",
																			"begin",
																			"if(rst_n_i = '0') then",
																			prefix.."_wr_sync0 <= '0';",
																			prefix.."_wr_sync1 <= '0';",
																			prefix.."_wr_sync2 <= '0';",
																			"elsif rising_edge("..field.clock..") then",
																			prefix.."_wr_sync0 <= "..prefix.."_wr_int;",
																			prefix.."_wr_sync1 <= "..prefix.."_wr_sync0;",
																			prefix.."_wr_sync2 <= "..prefix.."_wr_sync1;",
																			prefix.."_wr_o <= "..prefix.."_wr_sync1 and (not "..prefix.."_wr_sync2);",
																			"end if;",
																			"end process;" };
	end

end

-- generates VHDL code for single register field
function gen_vhdl_code_reg_field(field, reg)
  
	if(field.type == MONOSTABLE) then
    gen_vhdl_code_monostable(field, reg);
	elseif(field.type == BIT) then 
    gen_vhdl_code_bit(field, reg);
	elseif(field.type == SIGNED or field.type == UNSIGNED or field.type == SLV) then	    
    gen_vhdl_code_slv(field, reg);
	elseif(field.type == PASS_THROUGH) then
		gen_vhdl_code_passthrough(field, reg);
	end
end

-- generates VHDL for single register
function gen_vhdl_code_reg(reg)
  foreach_subfield(reg, function(field, reg) gen_vhdl_code_reg_field(field, reg); end );
end

function gen_vhdl_block_select_bits()
  return "wb_addr_i("..(address_bus_width-1).." downto "..(address_bus_width - address_bus_select_bits)..")";
end

-- generates the entire VHDL register-bank access stuff
function gen_vhdl_regbank()
  local s;
  local ramcount = 0;

  vhdl_new_code();

  foreach_reg(function(reg) 
								if(reg.__type == TYPE_REG) then
	    						gen_vhdl_code_reg(reg);
								end
					    end );

	vhdl ("");
	vhdl ("begin");

  vhdl ("-- MAIN REGISTER BANK PROCESS --");
  vhdl ("process (wb_clk_i, rst_n_i)");
  vhdl ("begin");

-- generate reset code for the main process
  vhdl ("if(rst_n_i = '0') then");
  vhdl ("-- Resetting the signals --");
  vhdl ("");
    
  foreach_field(function(field, reg) 
  								if (field.reset_code_main ~= nil) then 
								    vhdl_paste_code (field.reset_code_main); 
								  end
								end );

  vhdl ("wb_ack_regbank <= '0';");
  vhdl ("ack_in_progress <= '0';");    
  vhdl ("ack_cntr <= "..gen_vhdl_bin_literal(0,4)..";");    
  vhdl ("wb_data_out_int <= (others => '0');");

  vhdl ("");
  vhdl ("elsif rising_edge(wb_clk_i) then");
  vhdl ("-- ACK signal generator --");
  vhdl ("if ack_in_progress = '1' then");
  vhdl ("if(ack_cntr = "..gen_vhdl_bin_literal(0,4)..") then");

  foreach_reg(function(reg) 
								if(reg.__type == TYPE_REG) then
	    						foreach_subfield(reg, function(field, reg) 
	    							vhdl_paste_code(field.ackgen_code); 
	    						end );
								end
					    end );

  vhdl ("ack_in_progress <= '0';");
  vhdl ("wb_ack_regbank <= '0';");
  vhdl ("else");
  foreach_reg(function(reg) 
								if(reg.__type == TYPE_REG) then
	    						foreach_subfield(reg, function(field, reg) 
	    							vhdl_paste_code(field.ackgen_code_pre); 
	    						end );
								end
					    end );
  vhdl ("ack_cntr <= ack_cntr - 1;");
  vhdl ("if (ack_cntr = "..gen_vhdl_bin_literal(1,4)..") then");
 	vhdl ("wb_ack_regbank <= '1';");
	vhdl ("else");
 	vhdl ("wb_ack_regbank <= '0';");
	
  vhdl ("end if;");
  vhdl ("end if;");
  vhdl ("else");

-- count the RAMs in the design
  foreach_reg(function(reg) 
  							if(reg.__type == TYPE_RAM) then 
  								ramcount = ramcount + 1; 
  							end
  						end 
  						);
    
  if(ramcount > 0) then
		vhdl ("case "..gen_vhdl_block_select_bits().." is ");

		foreach_reg(function(reg) 
									if (reg.__type == TYPE_RAM) then
    								vhdl ("when "..gen_vhdl_bin_literal(reg.select_bits, address_bus_select_bits).." => ");
   									vhdl ("wb_ack_regbank <= '1'; ack_in_progress <= '1'; ack_cntr <= "..gen_vhdl_bin_literal(0,4).."; ");
									end
							 	end );
    
		vhdl ("when "..gen_vhdl_bin_literal(0, address_bus_select_bits).." => ");
  end 

    
  vhdl ("if ( wb_sel_i = '1' and wb_cyc_i = '1' and wb_stb_i = '1' ) then");

  if(regbank_address_bits > 0) then
   	vhdl ("case (wb_addr_i("..(regbank_address_bits -1 ).." downto 0)) is");
  end
    
  foreach_reg(function(reg) 
		if(reg.__type == TYPE_REG) then
	    local acklen = find_max(reg, "acklen");

	   	if(regbank_address_bits > 0) then
	   		vhdl ("when "..gen_vhdl_bin_literal(reg.base, regbank_address_bits).." => \t-- "..reg.name.."");
	   	end
		   	
			vhdl ("if(wb_we_i = '1') then -- write access");
			foreach_subfield(reg, function(field, reg) vhdl_paste_code(field.write_code); end );
			vhdl ("else -- read access");
			foreach_subfield(reg, function(field, reg) vhdl_paste_code(field.read_code); end );		
			vhdl ("end if; ");
		
		-- emit the ACK generation code. Special case for 1-cycle ACK
				--    if(acklen == 1) then
				--	vhdl("ack_in_progress <= '0'; ");
						--else
						
			vhdl("wb_ack_regbank <= "..csel((acklen==1),"'1'","'0'")..";");
			vhdl("ack_cntr <= "..gen_vhdl_bin_literal(math.max(acklen-1, 0),4)..";");
			vhdl("ack_in_progress <= '1'; ");
	  end
	end );

  if(regbank_address_bits > 0) then
		vhdl("when others => ");
		vhdl("ack_in_progress <= '0'; ack_cntr <= "..gen_vhdl_bin_literal(0,4).."; wb_ack_regbank <= '0';");	    
    vhdl ("end case;");
	end

  vhdl ("else");
	vhdl ("wb_ack_regbank <= '0';");
	vhdl ("end if;");

  if(ramcount > 0) then
		vhdl("when others => ack_in_progress <= '0'; ack_cntr <= "..gen_vhdl_bin_literal(0,4).."; ");	
--    vhdl_indent_left();	
 --  vhdl_indent_left();	
    vhdl("end case; ");
  end

  vhdl ("end if;") 
  vhdl ("end if;") 
  vhdl ("end process;");
    
  vhdl("");

--	vhdl_indent_right();
  foreach_reg(function(reg) 
								if(reg.extra_code ~= nil) then
									vhdl("-- extra code for reg/fifo/mem: "..reg.name);
									vhdl_paste_code(reg.extra_code);
    	 	  				vhdl("");
								end
						   foreach_subfield(reg, function(field, reg) vhdl_paste_code(field.extra_code); vhdl("");end );
							end);
    
	return code_vhdl;
end

